import importlib
import sys
import pathlib

"""
This logic is here to allow for relative imports via the .git repo,
as well as installed versions as a backup.
"""
if pathlib.Path('./glesys/__init__.py').absolute().exists():
	spec = importlib.util.spec_from_file_location("glesys", "./glesys/__init__.py")
	glesys = importlib.util.module_from_spec(spec)
	sys.modules["glesys"] = glesys
	spec.loader.exec_module(sys.modules["glesys"])
else:
	import glesys

if __name__ == '__main__':
	glesys.run_as_a_module()


import argparse
# https://towardsdatascience.com/dynamically-add-arguments-to-argparse-python-patterns-a439121abc39
args = None

parser = argparse.ArgumentParser("glesys")
parser.add_argument(
	"--verbose",
	required=False,
	action="store_true",
	default=False,
	help="Increase verbosity",
)

subparsers = parser.add_subparsers(help="Sub-commands help")

def load_arguments():
	global args

	args, unknowns = parser.parse_known_args()
	return args


import json
import pathlib
from .cli.args import args
from .output import log
from .models.configuration import Configuration

"""
Configuration will be loaded from a few different locations.
./glesys.toml has priority over /etc/glesys/glesys.toml but
--config will override any other location.
"""

config_location = None
if 'config' in dir(args) and args.config:
	if args.config.exists() is False:
		raise PermissionError(f"Could not load --conf from {args.config}, file does not exist or permission error.")
	log(f"Loading default configuration from --conf {args.config}", fg="gray")
	config_location = args.config
elif (path := pathlib.Path('./glesys.toml')).exists():
	config_location = path
elif (path := pathlib.Path('/etc/glesys/glesys.toml')).exists():
	config_location = path

try:
	# Py3.10+
	import tomllib
	file_mode = 'rb'
except:
	# Py3.9 or below (external dep needed)
	import toml as tomllib
	file_mode = 'r'

with config_location.open(file_mode) as f:
	configuration = Configuration(**tomllib.load(f))


import pydantic
import urllib.request
import json
import base64
import logging
import sys

from ..output import log


def get_request(endpoint):
	from ..session import configuration

	credentials = base64.b64encode(f"{configuration.credentials.user}:{configuration.credentials.key}".encode())

	req = urllib.request.Request(f"{configuration.network.host}{endpoint}")
	req.add_header('Accept', 'application/json; charset=utf-8')
	req.add_header('Authorization', f'Basic {credentials.decode()}')

	response = urllib.request.urlopen(req)

	return response.read()


def post_request(endpoint, payload):
	from ..session import configuration

	credentials = base64.b64encode(f"{configuration.credentials.user}:{configuration.credentials.key}".encode())

	req = urllib.request.Request(f"{configuration.network.host}{endpoint}")
	req.add_header('Content-Type', 'application/json; charset=utf-8')
	req.add_header('Authorization', f'Basic {credentials.decode()}')

	json_payload = json.dumps(payload).encode('utf-8')
	
	req.add_header('Content-Length', len(json_payload))
	response = urllib.request.urlopen(req, json_payload)

	response_data = json.loads(response.read().decode())

	return response_data


class DNS(pydantic.BaseModel):
	def all_domains(self):
		from ..session import configuration
		
		log(f"Listing all domains for user {configuration.credentials.user}", fg="gray", level=logging.DEBUG)
		response = get_request('/domain/list/')

		data = json.loads(response.decode())

		if data.get('response', {}).get('status', {}).get('code') == 200:
			return data.get('response', {}).get('domains', [])

	def all_records(self, domain):
		log(f"Listing all records for domain: {domain}", fg="gray", level=logging.DEBUG)
		response = post_request('/domain/listrecords/', {'domainname' : domain})

		if response.get('response', {}).get('status', {}).get('code') == 200:
			return response.get('response', {}).get('records', [])

	def find(self, domain, record=None, type=None, data=None):
		log(f"Performing a search on domain {domain} using query record={record}, type={type}, data={data}", fg="gray", level=logging.DEBUG)
		for dns_entry in self.all_records(domain):
			if all([
				record is None or record == dns_entry.get('host'),
				type is None or type == dns_entry.get('type'),
				data is None or data in dns_entry.get('data')
			]):
				yield dns_entry

	def add_record(self, domain, record):
		if '--verbose' in sys.argv:
			log(f"Attempting to add a record in domain {domain} with the data {record}", fg="gray", level=logging.INFO)

		record_data = {
			**{'ttl' : 360, 'domainname' : domain},
			**record
		}

		if (
			response := post_request('/domain/addrecord', record_data)
		).get('response', {}).get('status', {}).get('code') != 200:
			raise PermissionError(f"Could not create record {record_data} due to: {response}")

		return True

	def delete_record(self, domain, record):
		if '--verbose' in sys.argv:
			log(f"Attempting to delete all records in domain {domain} matching {record}", fg="gray", level=logging.INFO)

		for dns_entry in self.find(domain=domain, record=record.get('host'), type=record.get('type'), data=record.get('data')):
			log(f"Deleting record {dns_entry}", fg="gray", level=logging.DEBUG)

			if (response := post_request('/domain/deleterecord', {
				'recordid' : dns_entry.get('recordid')
			}).get('response', {}).get('status', {}).get('code')) != 200:
				raise PermissionError(f"Could not delete record {dns_entry} due to: {response}")

		return True


class API(pydantic.BaseModel):
	user :str
	key :str
	host :str

	dns :DNS = DNS()


import json
import logging
import datetime
import time

from ...output import log, stylize_output
from ...session import api, configuration
from ...workers import SysCommandWorker

def has_expired(date):
	
	expiry_date = datetime.datetime.fromtimestamp(
		time.mktime(
			time.strptime(date, '%Y-%m-%d')
		)
	)

	if (expiry_date - datetime.datetime.now()).total_seconds() > 0:
		return False

	return True


def letsencrypt_entrypoint(args, **kwargs):
	if args.all_domains:
		certbot_domain_string = ""

		log(f"Fetching all domains at Glesys accessible by account {configuration.credentials.user}", fg="blue", level=logging.INFO)

		debug_output_domains = 0
		for domain in api.dns.all_domains():
			if domain.get('usingglesysnameserver') == "no":
				log(f"Ignoring domain {domain['domainname']} - not using Glesys nameserver", fg="gray", level=logging.INFO)
				continue

			if has_expired(domain.get('registrarinfo', {}).get('expire')):
				log(f"Ignoring domain {domain['domainname']} - has expired", fg="gray", level=logging.DEBUG)
				continue

			certbot_domain_string += f"*.{domain['domainname']},{domain['domainname']},"
			debug_output_domains += 1

		log(f"Requesting new certificates using LetsEncrypt for {debug_output_domains} domains", fg="blue", level=logging.INFO)


		certbot_cmd_string = "certbot certonly --non-interactive"
		if args.verbose:
			certbot_cmd_string += '--verbose'
		certbot_cmd_string += " --authenticator dns-glesys"
		certbot_cmd_string += " --preferred-challenges dns"
		certbot_cmd_string += f" --work-dir {args.workdir.resolve()}"
		certbot_cmd_string += f" --logs-dir {args.logsdir.resolve()}"
		certbot_cmd_string += f" --config-dir {args.confdir.resolve()}"
		certbot_cmd_string += " --text --agree-tos --email anton@hvornum.se"
		certbot_cmd_string += " --expand --renew-by-default"
		certbot_cmd_string += " --manual-auth-hook /tmp/exit.sh"
		if args.production:
			certbot_cmd_string += " --server https://acme-v02.api.letsencrypt.org/directory"
		else:
			certbot_cmd_string += " --dry-run"
			certbot_cmd_string += " --server https://acme-staging-v02.api.letsencrypt.org/directory"
		certbot_cmd_string += f" -d '{certbot_domain_string[:-1]}'"

		with open('/tmp/exit.sh', 'w') as fh:
			fh.write('#!/bin/bash\nexit(0)\n')

		# chmod +x /tmp/exit.sh
		worker = SysCommandWorker(certbot_cmd_string, peek_output=True)

		while worker.is_alive():
			time.sleep(1)


		log(f"Certificates have been generated successfully", fg="green", level=logging.INFO)


from .cli.args_router import loaded
from .cli.args import load_arguments

session = None
__version__ = "0.1"

def run_as_a_module():
	# global session

	args = load_arguments()
	
	# import importlib
	# import sys
	# import pathlib

	# Load .git version before the builtin version
	# if pathlib.Path('./glesys/session.py').absolute().exists():
	# 	spec = importlib.util.spec_from_file_location("session", "./glesys/session.py")
	# 	session = importlib.util.module_from_spec(spec)
	# 	sys.modules["session"] = session
	# 	spec.loader.exec_module(sys.modules["session"])
	# else:
	# 	import glesys.session
	
	if 'func' in dir(args):
		args.func(args)


import pathlib
from ..args import subparsers
from ...dns.cli_handler import dns_entrypoint

parser_dns = subparsers.add_parser("dns", help="Perform DNS actions on Glesys API")
parser_dns.add_argument(
	"--color",
	required=False,
	action="store_true",
	default=False,
	help="Adds colors to certain elements",
)

parser_dns.add_argument(
	"--all-domains",
	required=False,
	action="store_true",
	default=False,
	help="List all domains registered via Glesys",
)
parser_dns.add_argument(
	"--all-records",
	required=False,
	action="store_true",
	default=False,
	help="List all records for a given --domain",
)
parser_dns.add_argument(
	"--find",
	required=False,
	action="store_true",
	default=False,
	help="Search through a given --domain for --record, --type or --data",
)
parser_dns.add_argument(
	"--add",
	required=False,
	action="store_true",
	default=False,
	help="Adds a record in --domain using --record, --type, --data and -ttl",
)
parser_dns.add_argument(
	"--delete",
	required=False,
	action="store_true",
	default=False,
	help="Adds a record in --domain using --record, --type, --data and -ttl",
)

parser_dns.add_argument(
	"--domain",
	required=False,
	type=str,
	help="Defines which domain to work on",
)
parser_dns.add_argument(
	"--record",
	required=False,
	type=str,
	help="Filters which record name to look for during search",
)
parser_dns.add_argument(
	"--type",
	required=False,
	type=str,
	help="Filters which record type to look for during search",
)
parser_dns.add_argument(
	"--data",
	required=False,
	type=str,
	help="Filters which data/content to look for in the record search",
)
parser_dns.add_argument(
	"--ttl",
	required=False,
	type=int,
	help="Defines the TTL when using --add to create a DNS entry",
)
parser_dns.set_defaults(func=dns_entrypoint)


from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
	from .syscalls import SysCommandWorker

class RequirementError(BaseException):
	pass

class SysCallError(BaseException):
	def __init__(self, message :str, exit_code :Optional[int] = None, worker :Optional['SysCommandWorker'] = None) -> None:
		super(SysCallError, self).__init__(message)
		self.message = message
		self.exit_code = exit_code
		self.worker = worker


import json

from ..output import log, stylize_output
from ..session import api


def dns_entrypoint(args, **kwargs):
	if args.all_domains:
		result = json.dumps(api.dns.all_domains(), indent=4)
		if args.color:
			result = result.replace('"usingglesysnameserver": "no",', stylize_output('"usingglesysnameserver": "no",', fg="orange"))
			result = result.replace('"state": "EXPIRING60",', stylize_output('"state": "EXPIRING60",', fg="red"))
		print(result)
	elif args.all_records:
		print(json.dumps(api.dns.all_records(args.domain), indent=4))

	elif args.find:
		print(list(api.dns.find(domain=args.domain, record=args.record, type=args.type, data=args.data)))

	elif args.add:
		api.dns.add_record(args.domain, {
			'host': args.record,
			'data': args.data,
			'type': args.type,
			'ttl': args.ttl
		})

	elif args.delete:
		api.dns.delete_record(args.domain, {
			'host': args.record,
			'data': args.data,
			'type': args.type,
			'ttl': args.ttl
		})


"""DNS Authenticator for slimDNS."""
import json
import logging
import sys

import urllib
import zope.interface
import psycopg2, psycopg2.extras
from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT

from certbot import errors
from certbot import interfaces
from certbot.plugins import dns_common

from ...output import log
from ...session import api

@zope.interface.implementer(interfaces.IAuthenticator)
@zope.interface.provider(interfaces.IPluginFactory)
class Authenticator(dns_common.DNSAuthenticator):
	"""DNS Authenticator for ISPConfig

	This Authenticator uses the ISPConfig Remote REST API to fulfill a dns-01 challenge.
	"""

	description = ('Obtain certificates using a DNS TXT record.')
	ttl = 60

	def __init__(self, *args, **kwargs):
		super(Authenticator, self).__init__(*args, **kwargs)

	def _setup_credentials(self):
		# Just have to be here for certbot reasons
		return True

	def more_info(self): # pylint: disable=missing-docstring,no-self-use
		return 'This plugin configures a DNS TXT record to respond to a dns-01 challenge using ' + \
			   'the Glesys API DNS endpoint.'

	def _perform(self, domain, validation_name, validation):
		api.dns.add_record(domain, {
			'host': validation_name,
			'data': validation,
			'type': 'TXT',
			'ttl': self.ttl
		})

	def _cleanup(self, domain, validation_name, validation):
		api.dns.delete_record(domain, {
			'host': validation_name,
			'data': validation,
			'type': 'TXT',
			'ttl': self.ttl
		})


import ipaddress
from pydantic import BaseModel
from typing import Dict, Optional, List

class Credentials(BaseModel):
	user :str
	key :str

class Network(BaseModel):
	host :str = "https://api.glesys.com/"

class Configuration(BaseModel):
	credentials :Credentials
	network :Network


import pathlib
from ..args import subparsers
from ...hooks.letsencrypt.cli_handler import letsencrypt_entrypoint

parser_letsencrypt = subparsers.add_parser("lets-encrypt", help="Perform LetsEncrypt actions using Glesys")
parser_letsencrypt.add_argument(
	"--all-domains",
	required=False,
	action="store_true",
	default=False,
	help="Create certificate for all domains under the Glesys DNS hosting",
)
parser_letsencrypt.add_argument(
	"--individual",
	required=False,
	action="store_true",
	default=False,
	help="Divides up certificates per domain rather than unifying them all (preferred but not default)",
)
parser_letsencrypt.add_argument(
	"--production",
	required=False,
	action="store_true",
	default=False,
	help="Replaces acme-staging-v02 with LetsEncrypt prod server https://acme-v02.api.letsencrypt.org/directory",
)

parser_letsencrypt.add_argument(
	"--workdir",
	required=False,
	type=pathlib.Path,
	default=pathlib.Path('/var/lib/letsencrypt'),
	help="The path of the certbot working directory",
)
parser_letsencrypt.add_argument(
	"--logsdir",
	required=False,
	type=pathlib.Path,
	default=pathlib.Path('/var/log/letsencrypt'),
	help="The path of the certbot log directory",
)
parser_letsencrypt.add_argument(
	"--confdir",
	required=False,
	type=pathlib.Path,
	default=pathlib.Path('/etc/letsencrypt'),
	help="The path of the certbot config directory",
)

parser_letsencrypt.set_defaults(func=letsencrypt_entrypoint)


import logging
import os
import sys
from pathlib import Path
from typing import Dict, Union, List, Any, Callable, Optional

from dataclasses import asdict, is_dataclass


class Journald:
	@staticmethod
	def log(message :str, level :int = logging.DEBUG) -> None:
		try:
			import systemd.journal  # type: ignore
		except ModuleNotFoundError:
			return None

		log_adapter = logging.getLogger('archinstall')
		log_fmt = logging.Formatter("[%(levelname)s]: %(message)s")
		log_ch = systemd.journal.JournalHandler()
		log_ch.setFormatter(log_fmt)
		log_adapter.addHandler(log_ch)
		log_adapter.setLevel(logging.DEBUG)

		log_adapter.log(level, message)

# Found first reference here: https://stackoverflow.com/questions/7445658/how-to-detect-if-the-console-does-support-ansi-escape-codes-in-python
# And re-used this: https://github.com/django/django/blob/master/django/core/management/color.py#L12
def supports_color() -> bool:
	"""
	Return True if the running system's terminal supports color,
	and False otherwise.
	"""
	supported_platform = sys.platform != 'win32' or 'ANSICON' in os.environ

	# isatty is not always implemented, #6223.
	is_a_tty = hasattr(sys.stdout, 'isatty') and sys.stdout.isatty()
	return supported_platform and is_a_tty


# Heavily influenced by: https://github.com/django/django/blob/ae8338daf34fd746771e0678081999b656177bae/django/utils/termcolors.py#L13
# Color options here: https://askubuntu.com/questions/528928/how-to-do-underline-bold-italic-strikethrough-color-background-and-size-i
def stylize_output(text: str, *opts :str, **kwargs) -> str:
	"""
	Adds styling to a text given a set of color arguments.
	"""
	opt_dict = {'bold': '1', 'italic': '3', 'underscore': '4', 'blink': '5', 'reverse': '7', 'conceal': '8'}
	colors = {
		'black' : '0',
		'red' : '1',
		'green' : '2',
		'yellow' : '3',
		'blue' : '4',
		'magenta' : '5',
		'cyan' : '6',
		'white' : '7',
		'teal' : '8;5;109',      # Extended 256-bit colors (not always supported)
		'orange' : '8;5;208',    # https://www.lihaoyi.com/post/BuildyourownCommandLinewithANSIescapecodes.html#256-colors
		'darkorange' : '8;5;202',
		'gray' : '8;5;246',
		'grey' : '8;5;246',
		'darkgray' : '8;5;240',
		'lightgray' : '8;5;256'
	}
	foreground = {key: f'3{colors[key]}' for key in colors}
	background = {key: f'4{colors[key]}' for key in colors}
	reset = '0'

	code_list = []
	if text == '' and len(opts) == 1 and opts[0] == 'reset':
		return '\x1b[%sm' % reset

	for k, v in kwargs.items():
		if k == 'fg':
			code_list.append(foreground[str(v)])
		elif k == 'bg':
			code_list.append(background[str(v)])

	for o in opts:
		if o in opt_dict:
			code_list.append(opt_dict[o])

	if 'noreset' not in opts:
		text = '%s\x1b[%sm' % (text or '', reset)

	return '%s%s' % (('\x1b[%sm' % ';'.join(code_list)), text or '')


def log(*args :str, **kwargs :Union[str, int, Dict[str, Union[str, int]]]) -> None:
	string = orig_string = ' '.join([str(x) for x in args])

	# Attempt to colorize the output if supported
	# Insert default colors and override with **kwargs
	if supports_color():
		kwargs = {'fg': 'white', **kwargs}
		string = stylize_output(string, **kwargs)

	Journald.log(string, level=int(str(kwargs.get('level', logging.INFO))))

	# Finally, print the log unless we skipped it based on level.
	# We use sys.stdout.write()+flush() instead of print() to try and
	# fix issue #94
	if kwargs.get('level', logging.INFO) != logging.DEBUG:
		sys.stdout.write(f"{string}\n")
		sys.stdout.flush()


from __future__ import annotations

import hashlib
import json
import logging
import os
import secrets
import shlex
import subprocess
import stat
import string
import sys
import time
import re
import urllib.parse
import urllib.request
import urllib.error
import pathlib
from datetime import datetime, date
from typing import Callable, Optional, Dict, Any, List, Union, Iterator, TYPE_CHECKING

if sys.platform == 'linux':
	from select import epoll, EPOLLIN, EPOLLHUP
else:
	import select
	EPOLLIN = 0
	EPOLLHUP = 0

	class epoll():
		""" #!if windows
		Create a epoll() implementation that simulates the epoll() behavior.
		This so that the rest of the code doesn't need to worry weither we're using select() or epoll().
		"""
		def __init__(self) -> None:
			self.sockets: Dict[str, Any] = {}
			self.monitoring: Dict[int, Any] = {}

		def unregister(self, fileno :int, *args :List[Any], **kwargs :Dict[str, Any]) -> None:
			try:
				del(self.monitoring[fileno]) # noqa: E275
			except:
				pass

		def register(self, fileno :int, *args :int, **kwargs :Dict[str, Any]) -> None:
			self.monitoring[fileno] = True

		def poll(self, timeout: float = 0.05, *args :str, **kwargs :Dict[str, Any]) -> List[Any]:
			try:
				return [[fileno, 1] for fileno in select.select(list(self.monitoring.keys()), [], [], timeout)[0]]
			except OSError:
				return []

from .exceptions import RequirementError, SysCallError
from .output import log

def gen_uid(entropy_length :int = 256) -> str:
	return hashlib.sha512(os.urandom(entropy_length)).hexdigest()

def generate_password(length :int = 64) -> str:
	haystack = string.printable # digits, ascii_letters, punctiation (!"#$[] etc) and whitespace
	return ''.join(secrets.choice(haystack) for i in range(length))

def multisplit(s :str, splitters :List[str]) -> str:
	s = [s, ]
	for key in splitters:
		ns = []
		for obj in s:
			x = obj.split(key)
			for index, part in enumerate(x):
				if len(part):
					ns.append(part)
				if index < len(x) - 1:
					ns.append(key)
		s = ns
	return s

def locate_binary(name :str) -> str:
	for PATH in os.environ['PATH'].split(':'):
		for root, folders, files in os.walk(PATH):
			for file in files:
				if file == name:
					return os.path.join(root, file)
			break # Don't recurse

	raise RequirementError(f"Binary {name} does not exist.")

def clear_vt100_escape_codes(data :Union[bytes, str]):
	# https://stackoverflow.com/a/43627833/929999
	if type(data) == bytes:
		vt100_escape_regex = bytes(r'\x1B\[[?0-9;]*[a-zA-Z]', 'UTF-8')
	else:
		vt100_escape_regex = r'\x1B\[[?0-9;]*[a-zA-Z]'

	for match in re.findall(vt100_escape_regex, data, re.IGNORECASE):
		data = data.replace(match, '' if type(data) == str else b'')

	return data

def json_dumps(*args :str, **kwargs :str) -> str:
	return json.dumps(*args, **{**kwargs, 'cls': JSON})

class JsonEncoder:
	@staticmethod
	def _encode(obj :Any) -> Any:
		"""
		This JSON encoder function will try it's best to convert
		any archinstall data structures, instances or variables into
		something that's understandable by the json.parse()/json.loads() lib.

		_encode() will skip any dictionary key starting with an exclamation mark (!)
		"""
		if isinstance(obj, dict):
			# We'll need to iterate not just the value that default() usually gets passed
			# But also iterate manually over each key: value pair in order to trap the keys.

			copy = {}
			for key, val in list(obj.items()):
				if isinstance(val, dict):
					# This, is a EXTREMELY ugly hack.. but it's the only quick way I can think of to trigger a encoding of sub-dictionaries.
					val = json.loads(json.dumps(val, cls=JSON))
				else:
					val = JsonEncoder._encode(val)

				if type(key) == str and key[0] == '!':
					pass
				else:
					copy[JsonEncoder._encode(key)] = val
			return copy
		elif hasattr(obj, 'json'):
			# json() is a friendly name for json-helper, it should return
			# a dictionary representation of the object so that it can be
			# processed by the json library.
			return json.loads(json.dumps(obj.json(), cls=JSON))
		elif hasattr(obj, '__dump__'):
			return obj.__dump__()
		elif isinstance(obj, (datetime, date)):
			return obj.isoformat()
		elif isinstance(obj, (list, set, tuple)):
			return [json.loads(json.dumps(item, cls=JSON)) for item in obj]
		elif isinstance(obj, pathlib.Path):
			return str(obj)
		else:
			return obj

	@staticmethod
	def _unsafe_encode(obj :Any) -> Any:
		"""
		Same as _encode() but it keeps dictionary keys starting with !
		"""
		if isinstance(obj, dict):
			copy = {}
			for key, val in list(obj.items()):
				if isinstance(val, dict):
					# This, is a EXTREMELY ugly hack.. but it's the only quick way I can think of to trigger a encoding of sub-dictionaries.
					val = json.loads(json.dumps(val, cls=UNSAFE_JSON))
				else:
					val = JsonEncoder._unsafe_encode(val)

				copy[JsonEncoder._unsafe_encode(key)] = val
			return copy
		else:
			return JsonEncoder._encode(obj)

class JSON(json.JSONEncoder, json.JSONDecoder):
	"""
	A safe JSON encoder that will omit private information in dicts (starting with !)
	"""
	def _encode(self, obj :Any) -> Any:
		return JsonEncoder._encode(obj)

	def encode(self, obj :Any) -> Any:
		return super(JSON, self).encode(self._encode(obj))

class UNSAFE_JSON(json.JSONEncoder, json.JSONDecoder):
	"""
	UNSAFE_JSON will call/encode and keep private information in dicts (starting with !)
	"""
	def _encode(self, obj :Any) -> Any:
		return JsonEncoder._unsafe_encode(obj)

	def encode(self, obj :Any) -> Any:
		return super(UNSAFE_JSON, self).encode(self._encode(obj))


class SysCommandWorker:
	def __init__(
		self,
		cmd :Union[str, List[str]],
		callbacks :Optional[Dict[str, Any]] = None,
		peek_output :Optional[bool] = False,
		environment_vars :Optional[Dict[str, Any]] = None,
		logfile :Optional[None] = None,
		working_directory :Optional[str] = './',
		remove_vt100_escape_codes_from_lines :bool = True
	):
		if not callbacks:
			callbacks = {}

		if not environment_vars:
			environment_vars = {}

		if type(cmd) is str:
			cmd = shlex.split(cmd)

		cmd = list(cmd) # This is to please mypy
		if cmd[0][0] != '/' and cmd[0][:2] != './':
			# "which" doesn't work as it's a builtin to bash.
			# It used to work, but for whatever reason it doesn't anymore.
			# We there for fall back on manual lookup in os.PATH
			cmd[0] = locate_binary(cmd[0])

		self.cmd = cmd
		self.callbacks = callbacks
		self.peek_output = peek_output
		# define the standard locale for command outputs. For now the C ascii one. Can be overridden
		self.environment_vars = {**environment_vars}
		self.logfile = logfile
		self.working_directory = working_directory

		self.exit_code :Optional[int] = None
		self._trace_log = b''
		self._trace_log_pos = 0
		self.poll_object = epoll()
		self.child_fd :Optional[int] = None
		self.started :Optional[float] = None
		self.ended :Optional[float] = None
		self.remove_vt100_escape_codes_from_lines :bool = remove_vt100_escape_codes_from_lines

	def __contains__(self, key: bytes) -> bool:
		"""
		Contains will also move the current buffert position forward.
		This is to avoid re-checking the same data when looking for output.
		"""
		assert type(key) == bytes

		if (contains := key in self._trace_log[self._trace_log_pos:]):
			self._trace_log_pos += self._trace_log[self._trace_log_pos:].find(key) + len(key)

		return contains

	def __iter__(self, *args :str, **kwargs :Dict[str, Any]) -> Iterator[bytes]:
		for line in self._trace_log[self._trace_log_pos:self._trace_log.rfind(b'\n')].split(b'\n'):
			if line:
				if self.remove_vt100_escape_codes_from_lines:
					line = clear_vt100_escape_codes(line)

				yield line + b'\n'

		self._trace_log_pos = self._trace_log.rfind(b'\n')

	def __repr__(self) -> str:
		self.make_sure_we_are_executing()
		return str(self._trace_log)

	def __enter__(self) -> 'SysCommandWorker':
		return self

	def __exit__(self, *args :str) -> None:
		# b''.join(sys_command('sync')) # No need to, since the underlying fs() object will call sync.
		# TODO: https://stackoverflow.com/questions/28157929/how-to-safely-handle-an-exception-inside-a-context-manager

		if self.child_fd:
			try:
				os.close(self.child_fd)
			except:
				pass

		if self.peek_output:
			# To make sure any peaked output didn't leave us hanging
			# on the same line we were on.
			sys.stdout.write("\n")
			sys.stdout.flush()

		if len(args) >= 2 and args[1]:
			log(args[1], level=logging.DEBUG, fg='red')

		if self.exit_code != 0:
			raise SysCallError(f"{self.cmd} exited with abnormal exit code [{self.exit_code}]: {self._trace_log[-500:]}", self.exit_code, worker=self)

	def is_alive(self) -> bool:
		self.poll()

		if self.started and self.ended is None:
			return True

		return False

	def write(self, data: bytes, line_ending :bool = True) -> int:
		assert type(data) == bytes  # TODO: Maybe we can support str as well and encode it

		self.make_sure_we_are_executing()

		if self.child_fd:
			return os.write(self.child_fd, data + (b'\n' if line_ending else b''))

		return 0

	def make_sure_we_are_executing(self) -> bool:
		if not self.started:
			return self.execute()
		return True

	def tell(self) -> int:
		self.make_sure_we_are_executing()
		return self._trace_log_pos

	def seek(self, pos :int) -> None:
		self.make_sure_we_are_executing()
		# Safety check to ensure 0 < pos < len(tracelog)
		self._trace_log_pos = min(max(0, pos), len(self._trace_log))

	def peak(self, output: Union[str, bytes]) -> bool:
		if self.peek_output:
			if type(output) == bytes:
				try:
					output = output.decode('UTF-8')
				except UnicodeDecodeError:
					return False

			sys.stdout.write(str(output))
			sys.stdout.flush()

		return True

	def poll(self) -> None:
		self.make_sure_we_are_executing()

		if self.child_fd:
			got_output = False
			for fileno, event in self.poll_object.poll(0.1):
				try:
					output = os.read(self.child_fd, 8192)
					got_output = True
					self.peak(output)
					self._trace_log += output
				except OSError:
					self.ended = time.time()
					break

			if self.ended or (got_output is False and pid_exists(self.pid) is False):
				self.ended = time.time()
				try:
					wait_status = os.waitpid(self.pid, 0)[1]
					self.exit_code = os.waitstatus_to_exitcode(wait_status)
				except ChildProcessError:
					try:
						wait_status = os.waitpid(self.child_fd, 0)[1]
						self.exit_code = os.waitstatus_to_exitcode(wait_status)
					except ChildProcessError:
						self.exit_code = 1

	def execute(self) -> bool:
		import pty

		if (old_dir := os.getcwd()) != self.working_directory:
			os.chdir(str(self.working_directory))

		# Note: If for any reason, we get a Python exception between here
		#   and until os.close(), the traceback will get locked inside
		#   stdout of the child_fd object. `os.read(self.child_fd, 8192)` is the
		#   only way to get the traceback without losing it.

		self.pid, self.child_fd = pty.fork()

		# https://stackoverflow.com/questions/4022600/python-pty-fork-how-does-it-work
		if not self.pid:
			os.execve(self.cmd[0], list(self.cmd), {**os.environ, **self.environment_vars})

		else:
			# Only parent process moves back to the original working directory
			os.chdir(old_dir)

		self.started = time.time()
		self.poll_object.register(self.child_fd, EPOLLIN | EPOLLHUP)

		return True

	def decode(self, encoding :str = 'UTF-8') -> str:
		return self._trace_log.decode(encoding)


class SysCommand:
	def __init__(self,
		cmd :Union[str, List[str]],
		callbacks :Optional[Dict[str, Callable[[Any], Any]]] = None,
		start_callback :Optional[Callable[[Any], Any]] = None,
		peek_output :Optional[bool] = False,
		environment_vars :Optional[Dict[str, Any]] = None,
		working_directory :Optional[str] = './',
		remove_vt100_escape_codes_from_lines :bool = True):

		_callbacks = {}
		if callbacks:
			for hook, func in callbacks.items():
				_callbacks[hook] = func
		if start_callback:
			_callbacks['on_start'] = start_callback

		self.cmd = cmd
		self._callbacks = _callbacks
		self.peek_output = peek_output
		self.environment_vars = environment_vars
		self.working_directory = working_directory
		self.remove_vt100_escape_codes_from_lines = remove_vt100_escape_codes_from_lines

		self.session :Optional[SysCommandWorker] = None
		self.create_session()

	def __enter__(self) -> Optional[SysCommandWorker]:
		return self.session

	def __exit__(self, *args :str, **kwargs :Dict[str, Any]) -> None:
		# b''.join(sys_command('sync')) # No need to, since the underlying fs() object will call sync.
		# TODO: https://stackoverflow.com/questions/28157929/how-to-safely-handle-an-exception-inside-a-context-manager

		if len(args) >= 2 and args[1]:
			log(args[1], level=logging.ERROR, fg='red')

	def __iter__(self, *args :List[Any], **kwargs :Dict[str, Any]) -> Iterator[bytes]:
		if self.session:
			for line in self.session:
				yield line

	def __getitem__(self, key :slice) -> Optional[bytes]:
		if not self.session:
			raise KeyError(f"SysCommand() does not have an active session.")
		elif type(key) is slice:
			start = key.start if key.start else 0
			end = key.stop if key.stop else len(self.session._trace_log)

			return self.session._trace_log[start:end]
		else:
			raise ValueError("SysCommand() doesn't have key & value pairs, only slices, SysCommand('ls')[:10] as an example.")

	def __repr__(self, *args :List[Any], **kwargs :Dict[str, Any]) -> str:
		if self.session:
			return self.session._trace_log.decode('UTF-8', errors='backslashreplace')
		return ''

	def __json__(self) -> Dict[str, Union[str, bool, List[str], Dict[str, Any], Optional[bool], Optional[Dict[str, Any]]]]:
		return {
			'cmd': self.cmd,
			'callbacks': self._callbacks,
			'peak': self.peek_output,
			'environment_vars': self.environment_vars,
			'session': True if self.session else False
		}

	def create_session(self) -> bool:
		"""
		Initiates a :ref:`SysCommandWorker` session in this class ``.session``.
		It then proceeds to poll the process until it ends, after which it also
		clears any printed output if ``.peek_output=True``.
		"""
		if self.session:
			return self.session

		with SysCommandWorker(
			self.cmd,
			callbacks=self._callbacks,
			peek_output=self.peek_output,
			environment_vars=self.environment_vars,
			remove_vt100_escape_codes_from_lines=self.remove_vt100_escape_codes_from_lines,
			working_directory=self.working_directory) as session:

			if not self.session:
				self.session = session

			while self.session.ended is None:
				self.session.poll()

		if self.peek_output:
			sys.stdout.write('\n')
			sys.stdout.flush()

		return True

	def decode(self, fmt :str = 'UTF-8') -> Optional[str]:
		if self.session:
			return self.session._trace_log.decode(fmt)
		return None

	@property
	def exit_code(self) -> Optional[int]:
		if self.session:
			return self.session.exit_code
		else:
			return None

	@property
	def trace_log(self) -> Optional[bytes]:
		if self.session:
			return self.session._trace_log
		return None


def pid_exists(pid: int) -> bool:
	try:
		return any(subprocess.check_output(['/bin/ps', '--no-headers', '-o', 'pid', '-p', str(pid)]).strip())
	except subprocess.CalledProcessError:
		return False